"""Interval and IntervalSet

Represents an interval of time, and a sorted set of such intervals"""

import bisect

class IntervalError(Exception):
    """Error due to interval overlap, etc"""
    pass

class Interval(object):
    """Represents an interval of time."""

    start = None
    end = None

    def __init__(self, start, end):
        """
        'start' and 'end' are time, in floating point seconds since epoch.
        """
        if start > end:
            raise IntervalError("start must precede end")
        self.start = start
        self.end = end

    def __repr__(self):
        s = repr(self.start) + ", " + repr(self.end)
        return self.__class__.__name__ + "(" + s + ")"

    def __str__(self):
        return "[" + str(self.start) + " -> " + str(self.end) + "]"

    def __cmp__(self, other):
        """
        Compare two intervals.  If non-equal, order by start then end.
        """
        if self.start == other.start:
            if self.end < other.end:
                return -1
            if self.end > other.end:
                return 1
            return 0
        if self.start < other.start:
            return -1
        return 1
                
    def intersects(self, other):
        """Return True if two Interval objects intersect"""
        if (self.end <= other.start or
            self.start >= other.end):
            return False
        return True

    def subset(self, start, end):
        """Return a new Interval that is a subset of this one"""
        # A subclass that tracks additional data might override this.
        if start < self.start or end > self.end:
            raise IntervalError("not a subset")
        return Interval(start, end)

class IntervalSet(object):
    """
    A sorted, non-intersecting set of intervals.
    """

    def __init__(self, source=None):
        """
        'source' is an Interval or IntervalSet to add.
        """
        self.data = []
        if source is not None:
            if isinstance(source, Interval):
                self._add_single_interval(source)
            else:
                self._add_intervals(source)

    def __iter__(self):
        return self.data.__iter__()

    def __repr__(self):
        return self.__class__.__name__ + "(" + repr(self.data) + ")"

    def __str__(self):
        return  "[" + ", ".join([ str(x) for x in self.data ]) + "]"

    def __eq__(self, other):
        """Test equality of two IntervalSets.

        Treats adjacent Intervals as equivalent to one long interval,
        so this function really tests whether the IntervalSets cover
        the same spans of time."""
        i = 0
        j = 0
        outside = True

        def is_adjacent(a, b):
            """Return True if two Intervals are adjacent (same end or start)"""
            if a.end == b.start or b.end == a.start:
                return True
            else:
                return False

        try:
            while True:
                if (outside):
                    # To match, we need to be finished both sets
                    if (i >= len(self) and j >= len(other)):
                        return True
                    # Or the starts need to match
                    if (self[i].start != other[j].start):
                        return False
                    outside = False
                else:
                    # We can move on if the two interval ends match
                    if (self[i].end == other[j].end):
                        i += 1
                        j += 1
                        outside = True
                    else:
                        # Whichever ends first needs to be adjacent to the next
                        if (self[i].end < other[j].end):
                            if (not is_adjacent(self[i],self[i+1])):
                                return False
                            i += 1
                        else:
                            if (not is_adjacent(other[j],other[j+1])):
                                return False
                            j += 1
        except IndexError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        return self.data.__getitem__(key)

    def __iadd__(self, other):
        """Inplace add -- modifies self

        This throws an exception if the regions being added intersect."""
        if isinstance(other, Interval):
            self._add_single_interval(other)
        else:
            self._add_intervals(other)
        return self
        
    def __add__(self, other):
        """Add -- returns a new object

        This throws an exception if the regions being added intersect."""
        new = IntervalSet(self)
        new += IntervalSet(other)
        return new

    def __and__(self, other):
        """
        Compute a new IntervalSet from the intersection of two others

        Output intervals are built as subsets of the intervals in the
        first argument (self).
        """
        out = IntervalSet()

        if isinstance(other, IntervalSet):
            # We were given a set -- intersect with each interval inside
            for interval in other.data:
                out += self & interval
        else:
            # Intersecting with a just a single interval.
            # This loop could likely be optimized using a bisect.
            for this in self.data:
                # If there's any overlap, add the overlapping region
                if this.intersects(other):
                    out += this.subset(max(this.start, other.start),
                                       min(this.end, other.end))
        return out

    def intersects(self, other):
        """Return True if this IntervalSet intersects another"""
        return len(self & other) > 0

    def _add_intervals(self, iterable):
        """Add each Interval from an interable to this set"""
        for element in iter(iterable):
            self._add_single_interval(element)

    def _add_single_interval(self, interval):
        """Add one Interval to this set"""
        if self.intersects(interval):
            raise IntervalError("Tried to add overlapping interval "
                                "to this set")
        bisect.insort(self.data, interval)
